set CHERRY /usr/local/cherry

set maxcalls 3
set tryinterval 60
set modules_dir $CHERRY/bin/caller/callgen/modules

proc GetWeekSeconds {} {
  set tmp [clock format [clock seconds] -format "%u %k %M. %S."]
  return [expr { ( [lindex $tmp 0] - 1 ) * 86400 + [lindex $tmp 1] * 3600 + [lindex $tmp 2] * 60 + [lindex $tmp 3] }]
}

proc GetPendingCalls {howmuch} {
  global dbh

  puts -nonewline "Retrieving pending calls..."

  set ws [GetWeekSeconds]
  set statement "SELECT caller_spool.id, caller_spool.job_id, caller_spool.oper_id, caller_spool.client_id, caller_spool.number, caller_jobs.subject FROM caller_spool, caller_jobs, caller_periods WHERE caller_spool.tries_left > 0 AND caller_spool.status = 'PENDING' AND caller_periods.from_time < $ws AND caller_periods.till_time > $ws AND caller_spool.plan_id = caller_periods.plan_id AND caller_spool.last_call_time + caller_periods.interval < [clock seconds] AND caller_jobs.status = 'RUNNING' AND caller_spool.job_id = caller_jobs.id ORDER BY caller_spool.last_call_time + caller_periods.interval ASC LIMIT $howmuch"
  set sth [pg_exec $dbh $statement]
  set rows [pg_result $sth -numTuples]

  puts "$rows found"

  set result {}

  for {set row 0} {$row < $rows} {incr row} {
    lappend result [pg_result $sth -getTuple $row]
  }

  pg_result $sth -clear

  return $result
}

proc Call {id job_id oper_id client_id number subject} {
  global dbh active_calls
  puts "Calling $number"
  set active_calls([MakeCall $number]) [list $id $job_id $number $subject $oper_id $client_id]

  pg_result [pg_exec $dbh "UPDATE caller_spool SET status = 'IN_CALL' WHERE id = $id"] -clear
  pg_result [pg_exec $dbh "INSERT INTO caller_log (oper_id, job_id, timestamp, client_id, number, event) VALUES ('$oper_id', $job_id, [clock seconds], '$client_id', '$number', 'Initiating call')"] -clear
}

proc LoadModule {module} {
  global modules_dir
  if [namespace exists $module] {
    return 1
  }

  if { ! [file exists $modules_dir/$module.tcl] } {
    puts "No module named $module found!"
    return 0
  }

  source $modules_dir/$module.tcl
  if { ! [namespace exists $module]} {
    puts "Invalid module $module"
    return 0
  }

  return 1
}

proc OnEvent {event token args} {
  puts "Event $event for connection $token with args $args appeared"
  switch $event {
    ConnectionEstablished {OnConnectionEstablished $token $args}
    ConnectionCleared {OnConnectionCleared $token $args}
    AnswerCall {OnAnswerCall $token $args}
    UserInput {OnUserInput $token $args}
    AudioEnd {OnAudioEnd $token}
  }
}

proc OnUserInput {token args} {
  global active_calls
  set value [lindex $args 0]
  puts "User at $token sent $value"

  foreach {id job_id number subject oper_id client_id} $active_calls($token) {}
  if [LoadModule $subject] {${subject}::OnUserInput $token $number $oper_id $client_id}
}

proc OnConnectionEstablished {token arglist} {
  global active_calls calling caller dbh
  set remoteParty [lindex $arglist 0]
  puts "We have a connection $token with $remoteParty!"

  foreach {id job_id number subject oper_id client_id} $active_calls($token) {}

  puts "$subject for $number $oper_id $client_id"
  if [LoadModule $subject] {${subject}::OnConnectionEstablished $token $number $oper_id $client_id}
  pg_result [pg_exec $dbh "INSERT INTO caller_log (oper_id, job_id, timestamp, client_id, number, event) VALUES ('$oper_id', $job_id, [clock seconds], '$client_id', '$number', 'Connection established')"] -clear
}

proc OnConnectionCleared {token arglist} {
  global active_calls dbh

  foreach {remoteParty reason} $arglist {}
  puts "Connection $token with $remoteParty cleared by $reason"
  foreach {id job_id number subject oper_id client_id} $active_calls($token) {}
  unset active_calls($token)

  pg_result [pg_exec $dbh "INSERT INTO caller_log (oper_id, job_id, timestamp, client_id, number, event) VALUES ('$oper_id', $job_id, [clock seconds], '$client_id', '$number', 'Connection cleared: $reason')"] -clear

  if { [string equal $reason TemporaryFailure] || [string equal $reason Q931Cause] } {
    pg_result [pg_exec $dbh "UPDATE caller_spool SET status = 'PENDING' WHERE id = $id"] -clear
    return
  }

  set result 0

  if [LoadModule $subject] {set result [${subject}::OnConnectionCleared $token $number $oper_id $client_id]}

  if $result {
    set status DONE
  } else {
    set status PENDING
  }

  puts "Status is $status for $number"

  pg_result [pg_exec $dbh "INSERT INTO caller_log (oper_id, job_id, timestamp, client_id, number, event) VALUES ('$oper_id', $job_id, [clock seconds], '$client_id', '$number', 'New status: $status')"] -clear


  set sth [pg_exec $dbh "UPDATE caller_spool SET status = '$status', tries_left = tries_left - 1, last_call_time = [clock seconds] WHERE id = $id"]

  if [string equal $status DONE] {
    # close done jobs
    pg_result [pg_exec $dbh "UPDATE caller_jobs SET status = 'DONE' WHERE (SELECT count(*) FROM caller_spool WHERE caller_spool.job_id = caller_jobs.id AND status != 'DONE') = 0"] -clear
  }

  CallPending
  pg_result $sth -clear
}

proc OnAnswerCall {token arglist} {
  foreach {caller dest number} $arglist {}
  puts "$caller \[$number\] calls us \[$dest\]"
  return Now
}

proc OnAudioEnd {token} {
  global active_calls
  puts "Finished playing for $token"

  foreach {id job_id number subject oper_id client_id} $active_calls($token) {}
  if [LoadModule $subject] {${subject}::OnAudioEnd $token $number $oper_id $client_id}
}

proc CallPending {} {
  global maxcalls active_calls
  set pending [GetPendingCalls [expr $maxcalls - [array size active_calls]]]

  foreach {call} $pending {
    foreach {id job_id oper_id client_id number subject} $call {}
    Call $id $job_id $oper_id $client_id $number $subject
   }
}

proc Periodic {} {
  global tryinterval
  CallPending
  after [expr $tryinterval * 1000] Periodic
}

# here we go
array set active_calls {}

load libpgtcl[info sharedlibextension]

set dbh [pg_connect -conninfo {dbname=cherry user=cherry}]

# flush invalid "IN_CALL" entries
pg_result [pg_exec $dbh "UPDATE caller_spool SET status = 'PENDING' WHERE status = 'IN_CALL'"] -clear

Periodic
