#!/usr/bin/perl
# SAF16E1 monitoring agent


use strict;
use Socket;
use Fcntl;

use SNMP_Session;
use SNMP_util;
use Data::Dumper;

require "/etc/cherry/monitoring/a4400_agent.cfg";

# NSC server
my $nsc_server = $a4400::nsc_server ? $a4400::nsc_server : "127.0.0.1:3456";


# how often we check switch parameters
my $switch_scan_period = $a4400::switch_scan_period ? $a4400::switch_scan_period : 300;

# maximum sleep time, should be shorter then other periods
my $sleep_time = $a4400::sleep_time ? $a4400::sleep_time : 5;

my $max_update_period = $a4400::max_update_period ? $a4400::max_update_period : 60;

# how long our data stays recent (if we won't send recency update
# during this period, data will be considered lost)
# must be longer than max_update_period
my $recency_time = $a4400::recency_time ? $a4400::recency_time : 120;

my $global_time = time;
my $last_update = $global_time - $max_update_period;
my $last_switch_scan = $global_time - $switch_scan_period;
my $logfile = $a4400::logfile ? $a4400::logfile : "/var/log/cherry/a4400.log";

# log procedure
sub logmsg {
        print time.":".(scalar localtime).": @_\n";

        if(-f $logfile) {
                open(LOG, ">>$logfile");
                print LOG time.":".(scalar localtime).": @_\n";
                close(LOG);
        }
}

my $buffer;

# autoflush
$| = 1;

my($remote, $port) = split(":", $nsc_server);

my $iaddr = inet_aton($remote) || die "no host: $remote";
my $paddr = sockaddr_in($port, $iaddr);
my $proto = getprotobyname('tcp');

socket(SOCK, PF_INET, SOCK_STREAM, $proto) || die "socket: $!";

connect(SOCK, $paddr) || die "connect: $!";

# autoflush on SOCK
my $old_fh = select;
select(SOCK);
$| = 1;
select($old_fh);

logmsg "connected to server";

sub switch_scan;

$SIG{PIPE} = sub {
	logmsg "NSC server died, shutting down";
	close SOCK;
	exit;
};

while(1) {
	if(time - $last_update >= $max_update_period) {
		# send update
		logmsg "sending update";

		foreach my $ref (values %a4400::objects) {
			my $path = $ref->{path};
			send_value($path);
		}

		$last_update = time; 
	}

	if(time - $last_switch_scan >= $switch_scan_period) {
		# do switch scan
		logmsg "scanning switch values";

		switch_scan;

		$last_switch_scan = time;
	}

	sleep($sleep_time);

	$global_time = time;
}

sub send_value {
	my($path, $type, $value, $e_level) = @_;

	my $cmd = defined $e_level ? "SET $path $recency_time -e$e_level $type $value" : "SET $path $recency_time $type $value";

	print SOCK "$cmd\n";

	logmsg $cmd;
}

sub switch_cmd {
	my ($host, $login, $password, $cmd) = @_;
	my $data;

	open(IN, "( sleep 3; echo '$login' ; sleep 2; echo '$password' ; sleep 2; echo '$cmd' ; sleep 10 ) | /usr/bin/telnet '$host' |");
	while(<IN>) {
		chop;
		$data .= "\n".$_; 
	} 

	return $data;
}

sub switch_scan {
	foreach my $host (keys %a4400::objects) {
		my $attrs = $a4400::objects{$host};
		my $path = $attrs->{path};
		my $login = $attrs->{login};
		my $password = $attrs->{password};
		my %ibs_state;
		my %ibs_busy;

		foreach my $ibs (@{$attrs->{ibs}}) {
			my ($crl, $cpl) = split('/',$ibs);
			logmsg "Checking IBS coupler: $crl $cpl";
			my $data = switch_cmd($host, $login, $password, "dectinfo 0 $crl $cpl");
			foreach my $ibs_line (grep (/For IBS/,split('\n', $data))) {
				my ($ibs_num, $busy) = $ibs_line =~ /For IBS([0-9\ ]*).*communications\ \=([0-9\ ]*)/;
				$ibs_busy{sprintf("%1d/%02d/%02d",$crl,$cpl,$ibs_num)} = $busy * 1;
			}
		}


		my $data = switch_cmd($host, $login, $password, "dectview ibs");

logmsg "$data";
		foreach my $ibs_line (grep (/RPN/,split('\n', $data))) {
			my ($ibs_num, $a,$a,$a,$a,$a, $state) = split(' ',$ibs_line);
			logmsg "$ibs_num = $state";
			$ibs_state{$ibs_num} = $state;
		}

		foreach my $ibs (keys %ibs_busy) {
			my $status;
			my $status_e;

			if($ibs_state{$ibs} =~ 'INSERV') {
				$status = 'up';
				$status_e = 0;
			} elsif($ibs_state{$ibs} =~ 'OUTSERV') {
				$status = 'down';
				$status_e = 2;
			} else {
				$status = 'unknown';
				$status_e = 1;
			}

			logmsg "IBS: $ibs: status = ($status, $status_e), busy = $ibs_busy{$ibs}";
			send_value("$path.ibs.$ibs.status", 'STATUS', $status, $status_e);
			send_value("$path.ibs.$ibs.busy", 'INTEGER', $ibs_busy{$ibs});
		}

	}
}

sub send_interface_status {
        my($host, $interface, $status) = @_;

        if(my $attrs = $a4400::objects{$host}) {
                my $path = $attrs->{path};
                        send_value("$interface.status", "STATUS", $status, $status eq 'down' ? 2 : ( $status
eq 'administratively down' ? undef: ( $status eq 'errors' ? 1 : 0)));
        }
}

