#!/usr/bin/perl
# Monitoring agent for CISCO Catalyst like devices
#
# see configuration file for more information

require "/etc/cherry/monitoring/catalyst_agent.cfg";

use strict;
use Socket;
use Fcntl;

use SNMP_Session;
use SNMP_util;
use Data::Dumper;

# NSC server
my $nsc_server = $catalyst::nsc_server ? $catalyst::nsc_server : "127.0.0.1:3456";

# how often we check snmp parameters
my $snmp_scan_period = $catalyst::snmp_scan_period ? $catalyst::snmp_scan_period : 300;

# maximum sleep time, should be shorter then other periods
my $sleep_time = $catalyst::sleep_time ? $catalyst::sleep_time : 5;

my $max_update_period = $catalyst::max_update_period ? $catalyst::max_update_period : 60;

# how long our data stays recent (if we won't send recency update
# during this period, data will be considered lost)
# must be longer than max_update_period
my $recency_time = $catalyst::recency_time ? $catalyst::recency_time : 120;

my $global_time = time;
my $last_update = $global_time - $max_update_period;
my $last_snmp_scan = $global_time - $snmp_scan_period;
my $logfile = $catalyst::logfile ? $catalyst::logfile : "/var/log/cherry/agent_catalyst.log";

# log procedure
sub logmsg {
        print time.":".(scalar localtime).": @_\n";

        if(-f $logfile) {
                open(LOG, ">>$logfile");
                print LOG time.":".(scalar localtime).": @_\n";
                close(LOG);
        }
}


my $buffer;

# send -HUP to syslogd
system("/usr/bin/killall -1 syslogd");

# we don't want to be blocked
fcntl(STDIN, F_SETFL, O_NONBLOCK);

# autoflush
$| = 1;

my $fd_set = '';

vec($fd_set, fileno(STDIN), 1) = 1;

my($remote, $port) = split(":", $nsc_server);

my $iaddr = inet_aton($remote) || die "no host: $remote";
my $paddr = sockaddr_in($port, $iaddr);
my $proto = getprotobyname('tcp');

socket(SOCK, PF_INET, SOCK_STREAM, $proto) || die "socket: $!";
while( !connect(SOCK, $paddr)) { logmsg "trying to connect to NSC..."; sleep 1;};

# autoflush on SOCK
my $old_fh = select;
select(SOCK);
$| = 1;
select($old_fh);

logmsg "connected to server";

sub parse_event;
sub snmp_scan;

$SIG{PIPE} = sub {
	logmsg "NSC server died, reconnecting...";
	while( 1 ) {
		my($remote, $port) = split(":", $nsc_server);
		my $iaddr = inet_aton($remote) || die "no host: $remote";
		my $paddr = sockaddr_in($port, $iaddr);
		my $proto = getprotobyname('tcp');
		close SOCK;
		socket(SOCK, PF_INET, SOCK_STREAM, $proto) || die "socket: $!";
		last if(connect(SOCK, $paddr));
		logmsg "cannot connect to NSC... $!"; 
		sleep 1;;
	}
	$old_fh = select;
	select(SOCK);
	$| = 1;
	select($old_fh);
	logmsg "connection to NSC has been restored";

	$last_update = 0;
	$last_snmp_scan = 0;
};

while(1) {
	my $rin;

	if(time - $last_update >= $max_update_period) {
		# send update
		logmsg "sending update: $last_update + $max_update_period = $global_time";

		foreach my $ref (values %catalyst::objects) {
			my $path = $ref->{path};
			send_value($path);
		}

		$last_update = time;
	}

	if(time - $last_snmp_scan >= $snmp_scan_period) {
		# do snmp scan
		logmsg "scanning snmp values";

		snmp_scan;

		$last_snmp_scan = time;
	}

	my $nfound = select($rin = $fd_set, undef, undef, $sleep_time);

	$global_time = time;

	if($nfound) {
		# check STDIN for log messages
		my($recv, $buf, $rc);

		while( $rc = sysread(STDIN, $buf, 1024) > 0) { $recv .= $buf }

		if($rc < 0 || length($recv) == 0) {
			logmsg "syslog error: $rc";
			sleep 1;
			next;
		}

		logmsg "got from syslog: $recv";

		$recv = $buffer.$recv;

		while($recv =~ /^(.*?)\n(.*)$/s) {
			my $event = $1;
			$recv = $2;

			logmsg "event: $event";

			parse_event($event);
		}

		$buffer = $recv;
	}
}

sub send_value {
        my($path, $type, $value, $e_level) = @_;

        my $cmd = defined $e_level ? "SET $path $recency_time -e$e_level $type $
value" : "SET $path $recency_time $type $value";

        print SOCK "$cmd\n";

        logmsg $cmd;
}

sub send_interface_status {
	my($host, $interface, $status) = @_;

	if(my $attrs = $catalyst::objects{$host}) {
		my $path = $attrs->{path};
#		my @interfaces = @{ $attrs->{interfaces} };

#		my $is_there;
#
#		foreach(@interfaces) {
#			if($_ eq $interface) {
#				$is_there = 1;
#				last;
#			}
#		}

#		if($is_there) {
			send_value("$path.interfaces.$interface.status", "STATUS", $status, $status eq 'down' ? 2 : ( $status eq 'administratively down' ? undef: ( $status eq 'errors' ? 1 : 0)));
#		}
	}
}

sub parse_event {
	my $event = shift;

	if(my($host, $interface, $state) = $event =~ /^\w+\s+\d+\s+\d+:\d+:\d+\s+([A-Za-z0-9\-\.]+)\s+\d+:\s+\w+:\s+\%LINK-4-ERROR:\s+([A-Za-z\/\d]+)\s+.*experiencing\s+([a-zA-Z0-9]+)$/) {

		logmsg "interface errors: $host, $interface, $state";
		send_interface_status($host, $interface, $state);
	}

	if(my($host, $interface, $state) = $event =~ /^\w+\s+\d+\s+\d+:\d+:\d+\s+([A-Za-z0-9\-\.]+)\s+\d+:\s+\w+:\s+\%LINK-3-UPDOWN:\s+Interface\s+([A-Za-z\/\d]+),\s+changed\s+state\s+to\s+(up|down)$/) {

		logmsg "interface up-down: $host, $interface, $state";
		send_interface_status($host, $interface, $state);
	}

	if(my($host, $interface, $state) = $event =~ /^\w+\s+\d+\s+\d+:\d+:\d+\s+([A-Za-z0-9\-\.]+)\s+\d+:\s+\w+:\s+\%LINK-5-CHANGED:\s+Interface\s+([A-Za-z\/\d]+),\s+changed\s+state\s+to\s+administratively\s+down$/) {

		logmsg "interface change: $host, $interface, administratively down";
		send_interface_status($host, $interface, 'administratively down');
	}
}

sub snmp_scan {

	foreach my $host (keys %catalyst::objects) {
		my $attrs = $catalyst::objects{$host};
		my $path = $attrs->{path};
		my $community = $attrs->{community};
		#my @interfaces = @{ $attrs->{interfaces} };
		my @names;
		my @admin_state;
		my @oper_state;
		my %indices;

#next if($host !~ /as5300/);

		my($if_num) = snmpget("$community\@$host", '.1.3.6.1.2.1.2.1.0');


		for (my $i = 1; $i < $if_num; $i++) {
			my ($name) = snmpget("$community\@$host",".1.3.6.1.2.1.2.2.1.2.$i");
			foreach (@{ $attrs->{interfaces} } ) {
				if($name eq $_) {
					$indices{$name} = $i;
					push @names, $name;

				}
			}
		}


		#my @names = snmpget("$community\@$host", map { ".1.3.6.1.2.1.2.2.1.2.$_"} 1..$if_num);

		my @interfaces = @names;


		foreach my $interface (@interfaces) {

			my ($oper_state) = snmpget("$community\@$host", ".1.3.6.1.2.1.2.2.1.8." . $indices{$interface} );
			push @oper_state, $oper_state;

			my ($admin_state) = snmpget("$community\@$host", ".1.3.6.1.2.1.2.2.1.7." . $indices{$interface} );
			push @admin_state, $admin_state;

			send_value("$path.interfaces.$interface.status", "STATUS", $admin_state == 2 ? ('administratively down', undef) : ( $oper_state == 1 ? ('up', 0) : ( $oper_state == 2 ? ('down', 2) : ('unknown', 1))));

			my($descr) = snmpget("$community\@$host", ".1.3.6.1.2.1.31.1.1.1.18." . $indices{$interface} );
			send_value("$path.interfaces.$interface.descr", "STRING", $descr);

			my($in_octets) = snmpget("$community\@$host", ".1.3.6.1.2.1.2.2.1.10." . $indices{$interface} );
			send_value("$path.interfaces.$interface.stats.in_octets", "BYTECOUNTER", $in_octets);

			my($out_octets) = snmpget("$community\@$host", ".1.3.6.1.2.1.2.2.1.16." . $indices{$interface} );
			send_value("$path.interfaces.$interface.stats.out_octets", "BYTECOUNTER", $out_octets);

			my $errors = snmpget("$community\@$host", ".1.3.6.1.2.1.2.2.1.14." . $indices{$interface} );
			send_value("$path.interfaces.$interface.stats.errors", "ERRORS", $errors);
		}
	}
}
