#!/usr/bin/perl
# Monitoring agent for CISCO Catalyst like devices
#
# see configuration file for more information

require "meridian_agent.cfg";

use strict;
use Socket;
use Fcntl;
use FileHandle;

use Data::Dumper;

# NSC server
my $nsc_server = $meridian::nsc_server ? $meridian::nsc_server : "127.0.0.1:3456";

# how often we check port parameters
my $port_scan_period = $meridian::port_scan_period ? $meridian::port_scan_period : 300;

# maximum sleep time, should be shorter then other periods
my $sleep_time = $meridian::sleep_time ? $meridian::sleep_time : 5;

my $max_update_period = $meridian::max_update_period ? $meridian::max_update_period : 60;

# how long our data stays recent (if we won't send recency update
# during this period, data will be considered lost)
# must be longer than max_update_interval
my $recency_time = $meridian::recency_time ? $meridian::recency_time : 120;

my $global_time = time;
my $last_update = $global_time - $max_update_period;
my $last_port_scan = $global_time - $port_scan_period;
my $logfile = $meridian::logfile ? $meridian::logfile : "/var/log/cherry/agent_meridian.log";

# log procedure
sub logmsg {
        print time.":".(scalar localtime).": @_\n";

        if(-f $logfile) {
                open(LOG, ">>$logfile");
                print LOG time.":".(scalar localtime).": @_\n";
                close(LOG);
        }
}


my $buffer;

# send -HUP to syslogd
# system("/usr/bin/killall -1 syslogd");

# we don't want to be blocked
fcntl(STDIN, F_SETFL, O_NONBLOCK);

# autoflush
$| = 1;

my $fd_set = '';

vec($fd_set, fileno(STDIN), 1) = 1;

my($remote, $port) = split(":", $nsc_server);

my $iaddr = inet_aton($remote) || die "no host: $remote";
my $paddr = sockaddr_in($port, $iaddr);
my $proto = getprotobyname('tcp');

socket(SOCK, PF_INET, SOCK_STREAM, $proto) || die "socket: $!";
while( !connect(SOCK, $paddr)) { logmsg "trying to connect to NSC..."; sleep 1;};

# autoflush on SOCK
my $old_fh = select;
select(SOCK);
$| = 1;
select($old_fh);

logmsg "connected to server";

sub parse_event;
sub port_scan;

$SIG{PIPE} = sub {
	logmsg "NSC server died, reconnecting...";
	while( 1 ) {
		my($remote, $port) = split(":", $nsc_server);
		my $iaddr = inet_aton($remote) || die "no host: $remote";
		my $paddr = sockaddr_in($port, $iaddr);
		my $proto = getprotobyname('tcp');
		close SOCK;
		socket(SOCK, PF_INET, SOCK_STREAM, $proto) || die "socket: $!";
		last if(connect(SOCK, $paddr));
		logmsg "cannot connect to NSC... $!"; 
		sleep 1;;
	}
	$old_fh = select;
	select(SOCK);
	$| = 1;
	select($old_fh);
	logmsg "connection to NSC has been restored";

	$last_update = 0;
	$last_port_scan = 0;
};

while(1) {
	my $rin;

	if(time - $last_update >= $max_update_period) {
		# send update
		logmsg "sending update: $last_update + $max_update_period = $global_time";

		foreach my $ref (values %meridian::objects) {
			my $path = $ref->{path};
			send_value($path);
		}

		$last_update = time;
	}

	if(time - $last_port_scan >= $port_scan_period) {
		logmsg "scanning port values";

		port_scan;

		$last_port_scan = time;
	}

	my $nfound = select($rin = $fd_set, undef, undef, $sleep_time);

	$global_time = time;

	if($nfound) {
		# check STDIN for log messages
		my($recv, $buf, $rc);

		while( $rc = sysread(STDIN, $buf, 1024) > 0) { $recv .= $buf }

		if($rc < 0 || length($recv) == 0) {
			logmsg "syslog error: $rc";
			sleep 1;
			next;
		}

		logmsg "got from syslog: $recv";

		$recv = $buffer.$recv;

		while($recv =~ /^(.*?)\n(.*)$/s) {
			my $event = $1;
			$recv = $2;

			logmsg "event: $event";

			parse_event($event);
		}

		$buffer = $recv;
	}
}

sub send_value {
        my($path, $type, $value, $e_level) = @_;

        my $cmd = defined $e_level ? "SET $path $recency_time -e$e_level $type $
value" : "SET $path $recency_time $type $value";

        print SOCK "$cmd\n";

        logmsg $cmd;
}

sub send_port_status {
	my($host, $port, $status) = @_;

	if(my $attrs = $meridian::objects{$host}) {
		my $path = $attrs->{path};
		send_value("$path.loops.$port.status", "STATUS", $status, $status eq 'down' ? 2 : ( $status eq 'administratively down' ? undef: ( $status eq 'errors' ? 1 : 0)));
	}
}

sub parse_event {
	my $event = shift;

	logmsg "syslog events currently not supported";
}

sub port_disconnect {
	my $host = shift @_;

	my $dev = $meridian::objects{$host}->{port};

	syswrite(COM, "LOGO\r", 5);
	close(COM);
	
	logmsg "port closed";
}

sub port_cmd {
	my $cmd = shift @_;
	my $timeout = shift @_;
	my $buf;

	# clear input buffer before entering command...

	syswrite(COM, "\r", 1);
	sleep 1;
	while( sysread(COM, $_, 1024) > 0) { }

	syswrite(COM, $cmd, length($cmd));

	while ( 1 ) {
		sleep $timeout;
		while( 1 ) { 
			my $rc = sysread(COM, $_, 1024);
			last if($rc < 1);
			$buf = $buf . $_;
		}
		last if($buf =~ />/ || $buf =~ /\./);
		syswrite(COM, "\r", 1);
	}


	return $buf;
}

sub port_data {
	my $port = shift @_;
	my $data;


	$data = port_cmd("****\r****\rLD 60\r", 1);
	$data = port_cmd("STAT $port\r\r", 2);

	return $data;
}

sub port_connect {
	my $host = shift @_;
	my $dev = $meridian::objects{$host}->{port};
	my $buf;

	if(!sysopen(COM,"$dev", 2)) {
		logmsg "cannot open port $dev for host $host: $!";
		return;
	}

	#COM->autoflush(1);
	fcntl(COM, F_SETFL, O_NONBLOCK);

	logmsg "port $dev for host $host opened";


	# trying to log in

	syswrite(COM, "****\r****\rLOGO\r", 15);
	while( sysread(COM, $_, 1024) > 0) {};

	syswrite(COM, "LOII\r", 5); 
	syswrite(COM, $meridian::objects{$host}->{password}."\r", length($meridian::objects{$host}->{password}) + 1);
	sleep 1;

	while( sysread(COM, $_, 1024) > 0) { $buf = $buf . $_; };
	### logmsg "GOT: $buf"; 

	if($buf =~ /LOGGED/) {
		logmsg "logged in successfully";
		return 1;
	} else {
		logmsg "cannot log in";
	}
}

sub port_scan {


	foreach my $host (keys %meridian::objects) {

		next if (!$host);

		my $attrs = $meridian::objects{$host};

		my $path = $attrs->{path};


		# trying to connect to host using serial port

		next if(!port_connect($host));

		# Scanning DTI LOOPS

		foreach my $dti_loop (@{ $attrs->{loops}->{dti} }) {

			my $data = port_data($dti_loop); 
			my $lck  = 0;
			my $idl  = 0;
			my $bus  = 0;
			my $mbs  = 0;
			my $oper_state = 2; # admin down by default
			

			next if(!$data =~ /LOOP/);

			$oper_state = 0 if($data =~ /LOOP.*ENBL/);

			foreach (split('\n',$data)) {

				if(/VCE/) {
					my($ts1, $state1, $ts2, $state2)  = $_ =~ 
						/CH ([0-9]+) - ([A-Z]+) .* CH ([0-9]+) - ([A-Z]+) /;	
					### print "L: $ts1, $state1, $ts2, $state2 \n";
					$lck++ if($state1 =~ /LCK/);
					$lck++ if($state2 =~ /LCK/);
					$idl++ if($state1 =~ /IDL/);
					$idl++ if($state2 =~ /IDL/);
					$bus++ if($state1 =~ /BUS/);
					$bus++ if($state2 =~ /BUS/);
					$mbs++ if($state1 =~ /MBS/);
					$mbs++ if($state2 =~ /MBS/);
				}
			}

			$oper_state = 1 if($mbs + $lck > 0);

			send_value("$path.loops.$dti_loop.status", "STATUS", $oper_state == 2 ? ('administratively down', undef) : ( $oper_state == 1 ? ('errors', 1) : ( $oper_state == 0 ? ('up', 0) : ('unknown', 1))));

			#send_value("$path.interfaces.$interface.descr", "STRING", $descr);

			send_value("$path.loops.$dti_loop.stats.idle", "GAUGE", $idl);
			send_value("$path.loops.$dti_loop.stats.busy", "GAUGE", $bus);
			send_value("$path.loops.$dti_loop.stats.lockout", "GAUGE", $lck);
			send_value("$path.loops.$dti_loop.stats.mbusy", "GAUGE", $mbs);
		}




		# Scanning PRI LOOPS

		foreach my $pri_loop (@{ $attrs->{loops}->{pri} }) {

			my $data = port_data($pri_loop); 
print $data."\n";
			my $lck  = 0;
			my $idl  = 0;
			my $bus  = 0;
			my $mbs  = 0;
			my $oper_state = 2; # admin down by default
			

			next if(!$data =~ /LOOP/);

			$oper_state = 0 if($data =~ /LOOP.*ENBL/);

			foreach (split('\n',$data)) {

				if(/TIE/) {
					my($ts1, $state1, $ts2, $state2)  = $_ =~ 
						/CH ([0-9]+) - ([A-Z]+) .* CH ([0-9]+) - ([A-Z]+) /;	
					### print "L: $ts1, $state1, $ts2, $state2 \n";
					$lck++ if($state1 =~ /LCK/);
					$lck++ if($state2 =~ /LCK/);
					$idl++ if($state1 =~ /IDL/);
					$idl++ if($state2 =~ /IDL/);
					$bus++ if($state1 =~ /BUS/);
					$bus++ if($state2 =~ /BUS/);
					$mbs++ if($state1 =~ /MBS/);
					$mbs++ if($state2 =~ /MBS/);
				}
			}

			$oper_state = 1 if($mbs + $lck > 0);

			send_value("$path.loops.$pri_loop.status", "STATUS", $oper_state == 2 ? ('administratively down', undef) : ( $oper_state == 1 ? ('errors', 1) : ( $oper_state == 0 ? ('up', 0) : ('unknown', 1))));

			#send_value("$path.interfaces.$interface.descr", "STRING", $descr);

			send_value("$path.loops.$pri_loop.stats.idle", "INTEGER", $idl);
			send_value("$path.loops.$pri_loop.stats.busy", "INTEGER", $bus);
			send_value("$path.loops.$pri_loop.stats.lockout", "INTEGER", $lck);
			send_value("$path.loops.$pri_loop.stats.mbusy", "INTEGER", $mbs);
		}


		port_disconnect();
	}
}
