#!/usr/bin/perl
# Monitoring agent for Cherry RADIUS server 
#
# see configuration file for more information

require "/etc/cherry/monitoring/radsrv_agent.cfg";

use strict;
use Socket;
use Fcntl;
use FileHandle;

use Data::Dumper;

# NSC server
my $nsc_server = $radsrv::nsc_server ? $radsrv::nsc_server : "127.0.0.1:3456";

my $rad_scan_period = $radsrv::port_scan_period ? $radsrv::rad_scan_period : 300;

# maximum sleep time, should be shorter then other periods
my $sleep_time = $radsrv::sleep_time ? $radsrv::sleep_time : 5;

my $max_update_period = $radsrv::max_update_period ? $radsrv::max_update_period : 60;

# how long our data stays recent (if we won't send recency update
# during this period, data will be considered lost)
# must be longer than max_update_interval
my $recency_time = $radsrv::recency_time ? $radsrv::recency_time : 120;

my $global_time = time;
my $last_update = $global_time - $max_update_period;
my $last_rad_scan = $global_time - $rad_scan_period;
my $logfile = $radsrv::logfile ? $radsrv::logfile : "/var/log/cherry/agent_radsrv.log";

# log procedure
sub logmsg {
        print time.":".(scalar localtime).": @_\n";

        if(-f $logfile) {
                open(LOG, ">>$logfile");
                print LOG time.":".(scalar localtime).": @_\n";
                close(LOG);
        }
}


my $buffer;

# send -HUP to syslogd
# system("/usr/bin/killall -1 syslogd");

# we don't want to be blocked
fcntl(STDIN, F_SETFL, O_NONBLOCK);

# autoflush
$| = 1;

my $fd_set = '';

vec($fd_set, fileno(STDIN), 1) = 1;

my($remote, $port) = split(":", $nsc_server);

my $iaddr = inet_aton($remote) || die "no host: $remote";
my $paddr = sockaddr_in($port, $iaddr);
my $proto = getprotobyname('tcp');

socket(SOCK, PF_INET, SOCK_STREAM, $proto) || die "socket: $!";
while( !connect(SOCK, $paddr)) { logmsg "trying to connect to NSC..."; sleep 1;};

# autoflush on SOCK
my $old_fh = select;
select(SOCK);
$| = 1;
select($old_fh);

logmsg "connected to server";

sub parse_event;
sub rad_scan;

$SIG{PIPE} = sub {
	logmsg "NSC server died, reconnecting...";
	while( 1 ) {
		my($remote, $port) = split(":", $nsc_server);
		my $iaddr = inet_aton($remote) || die "no host: $remote";
		my $paddr = sockaddr_in($port, $iaddr);
		my $proto = getprotobyname('tcp');
		close SOCK;
		socket(SOCK, PF_INET, SOCK_STREAM, $proto) || die "socket: $!";
		last if(connect(SOCK, $paddr));
		logmsg "cannot connect to NSC... $!"; 
		sleep 1;;
	}
	$old_fh = select;
	select(SOCK);
	$| = 1;
	select($old_fh);
	logmsg "connection to NSC has been restored";

	$last_update = 0;
	$last_rad_scan = 0;
};

while(1) {
	my $rin;

	if(time - $last_update >= $max_update_period) {
		# send update
		logmsg "sending update: $last_update + $max_update_period = $global_time";

		foreach my $ref (values %radsrv::objects) {
			my $path = $ref->{path};
			send_value($path);
		}

		$last_update = time;
	}

	if(time - $last_rad_scan >= $rad_scan_period) {
		logmsg "scanning radius values";

		rad_scan;

		$last_rad_scan = time;
	}

	my $nfound = select($rin = $fd_set, undef, undef, $sleep_time);

	$global_time = time;

	if($nfound) {
		# check STDIN for log messages
		my($recv, $buf, $rc);

		while( $rc = sysread(STDIN, $buf, 1024) > 0) { $recv .= $buf }

		if($rc < 0 || length($recv) == 0) {
			logmsg "syslog error: $rc";
			sleep 1;
			next;
		}

		logmsg "got from syslog: $recv";

		$recv = $buffer.$recv;

		while($recv =~ /^(.*?)\n(.*)$/s) {
			my $event = $1;
			$recv = $2;

			logmsg "event: $event";

			parse_event($event);
		}

		$buffer = $recv;
	}
}

sub send_value {
        my($path, $type, $value, $e_level) = @_;

        my $cmd = defined $e_level ? "SET $path $recency_time -e$e_level $type $
value" : "SET $path $recency_time $type $value";

        print SOCK "$cmd\n";

        logmsg $cmd;
}


sub parse_event {
	my $event = shift;

	logmsg "syslog events currently not supported";
}

sub rad_scan {


	foreach my $host (keys %radsrv::objects) {

		next if (!$host);

		my $attrs = $radsrv::objects{$host};

		my $path = $attrs->{path};



		next if(!open(USM,$attrs->{usm}));

		my %radusm;
		my %radnas;
		my $service_type;
		my $nas;
		my $name;
		my $session_begin;
		my $session_last;
		my $src;
		my $dst;
		my $i;

		while(<USM>) {
			chop;
			($name, $service_type, $_,$session_begin,$session_last,$src,$_,$dst,$_,$_,$_,$nas) = split(',');
			$nas =~ s/\./\-/g;
			my ($sec,$min,$hour,$mday,$mon,
				$year,$wday,$yday,$isdst) = localtime($session_begin);
			@radusm{$service_type} ++;		
			$radnas{$nas}->{$service_type} ++ ;
			$radnas{$nas}->{'stat'}->{$service_type}->{$radnas{$nas}->{$service_type}} =  
				sprintf("%s %02d:%02d:%02d %02d-%02d-%04d %d %s %s",$name, $hour, $min, $sec, $mday, $mon+1, $year+1900, $session_last - $session_begin, $src, $dst);
		}

		close(USM);

		foreach $service_type (keys %radusm) {
			send_value("$path.total.$service_type", "INTEGER", $radusm{$service_type} , 0);
		}

		my $i = 0;

		foreach $nas (keys %radnas) {
			foreach $service_type (keys %{$radnas{$nas}}) {
				next if($service_type eq 'stat');
				send_value("$path.nas.$nas.$service_type", "INTEGER", $radnas{$nas}->{$service_type} , 0);
				foreach $i (sort keys %{$radnas{$nas}->{'stat'}->{$service_type}}) {
					send_value("$path.nas.$nas.details.$service_type.$i", "STRING", $radnas{$nas}->{'stat'}->{$service_type}->{$i}, 0);
				}
			}
		}


		next if(!open(FLOW,$attrs->{flow}));

		my $flows = 0;

		while(<FLOW>) {
			$flows ++;
		}

		close(FLOW);

		send_value("$path.netflow.cached_netflows", "INTEGER", $flows, 0);

	}
}
