#!/usr/bin/perl
# SAF16E1 monitoring agent


use strict;
use Socket;
use Fcntl;

use SNMP_Session;
use SNMP_util;
use Data::Dumper;

require "saf16e1_agent.cfg";

# NSC server
my $nsc_server = $saf16e1::nsc_server ? $saf16e1::nsc_server : "127.0.0.1:3456";


# how often we check snmp parameters
my $snmp_scan_period = $saf16e1::snmp_scan_period ? $saf16e1::snmp_scan_period : 300;

# maximum sleep time, should be shorter then other periods
my $sleep_time = $saf16e1::sleep_time ? $saf16e1::sleep_time : 5;

my $max_update_period = $saf16e1::max_update_period ? $saf16e1::max_update_period : 60;

# how long our data stays recent (if we won't send recency update
# during this period, data will be considered lost)
# must be longer than max_update_period
my $recency_time = $saf16e1::recency_time ? $saf16e1::recency_time : 120;

my $global_time = time;
my $last_update = $global_time - $max_update_period;
my $last_snmp_scan = $global_time - $snmp_scan_period;
my $logfile = $saf16e1::logfile ? $saf16e1::logfile : "/var/log/cherry/saf16e1.log";

# log procedure
sub logmsg { print STDERR "$0 $$: @_ at ", scalar localtime, "\n" }

my $buffer;

# autoflush
$| = 1;

my($remote, $port) = split(":", $nsc_server);

my $iaddr = inet_aton($remote) || die "no host: $remote";
my $paddr = sockaddr_in($port, $iaddr);
my $proto = getprotobyname('tcp');

socket(SOCK, PF_INET, SOCK_STREAM, $proto) || die "socket: $!";

connect(SOCK, $paddr) || die "connect: $!";

# autoflush on SOCK
my $old_fh = select;
select(SOCK);
$| = 1;
select($old_fh);

logmsg "connected to server";

sub snmp_scan;

$SIG{PIPE} = sub {
	logmsg "NSC server died, shutting down";
	close SOCK;
	exit;
};

while(1) {
	if(time - $last_update >= $max_update_period) {
		# send update
		logmsg "sending update";

		foreach my $ref (values %saf16e1::objects) {
			my $path = $ref->{path};
			send_value($path);
		}

		$last_update = time; 
	}

	if(time - $last_snmp_scan >= $snmp_scan_period) {
		# do snmp scan
		logmsg "scanning snmp values";

		snmp_scan;

		$last_snmp_scan = time;
	}

	sleep($sleep_time);

	$global_time = time;
}

sub send_value {
	my($path, $type, $value, $e_level) = @_;

	my $cmd = defined $e_level ? "SET $path $recency_time -e$e_level $type $value" : "SET $path $recency_time $type $value";

	print SOCK "$cmd\n";

	logmsg $cmd;
}

sub snmp_scan {
	foreach my $host (keys %saf16e1::objects) {
		my $attrs = $saf16e1::objects{$host};
		my $path = $attrs->{path};
		my $community = $attrs->{community};

		my($temp) = snmpget("$community\@$host", '.1.3.6.1.4.1.7571.100.1.1.2.13.1.6.0');

		next if(!$temp);

		my($temp_status) = 'normal';
		my($temp_e) = 0;
		my($indoor_e) = 0;
		my($outdoor_e) = 0;


		$temp_status = 'warning' if($temp >= $attrs->{warning_temperature});
		$temp_status = 'critical' if($temp >= $attrs->{critical_temperature});
		$temp_e = 1 if($temp >= $attrs->{warning_temperature});
		$temp_e = 2 if($temp >= $attrs->{critical_temperature});

		send_value("$path.temperature", 'TEMPERATURE', $temp, $temp_e);
		send_value("$path.temperature.status", 'STATUS', $temp_status, $temp_e);
##$temp >= $attrs->{critical_temperature} ? 2 : ($temp >= $attrs->{warning_temperature} ? 1 : 0));

		my ($indoor_signal_level) = snmpget("$community\@$host", ".1.3.6.1.4.1.7571.100.1.1.2.13.3.11.0" );
		$indoor_e = 1 if($indoor_signal_level < -12);
		$indoor_e = 2 if($indoor_signal_level < -14);
		send_value("$path.indoor_signal_level", 'SIGNALLEVEL', $indoor_signal_level, $indoor_e);

		my ($outdoor_signal_level) = snmpget("$community\@$host", ".1.3.6.1.4.1.7571.100.1.1.2.13.3.10.0" );
		$outdoor_e = 1 if($outdoor_signal_level < -65);
		$outdoor_e = 2 if($outdoor_signal_level < -71);
		send_value("$path.outdoor_signal_level", 'SIGNALLEVEL', $outdoor_signal_level, $outdoor_e);

		my @los = snmpget("$community\@$host", map { ".1.3.6.1.4.1.7571.100.1.1.2.13.4.$_.1.0" } 1..16 );
		my @ais = snmpget("$community\@$host", map { ".1.3.6.1.4.1.7571.100.1.1.2.13.4.$_.2.0" } 1..16 );

		foreach my $channel (1..16) {
			my $los = shift @los ? 'on' : 'off';
			my $ais = shift @ais ? 'on' : 'off';
			my $status = 'up';
			send_value("$path.channels.ch$channel.los", 'STATE', $los);
			send_value("$path.channels.ch$channel.ais", 'STATE', $ais);
			if($los eq 'on' || $ais eq 'on') { $status = 'down'; }
			if($los eq 'on' && $ais eq 'on') { $status = 'administratively down'; }
			send_interface_status($host, $path.".channels.ch".$channel, $status);
		}

	}
}

sub send_interface_status {
        my($host, $interface, $status) = @_;

        if(my $attrs = $saf16e1::objects{$host}) {
                my $path = $attrs->{path};
                        send_value("$interface.status", "STATUS", $status, $status eq 'down' ? 2 : ( $status
eq 'administratively down' ? undef: ( $status eq 'errors' ? 1 : 0)));
        }
}

