#!/usr/bin/perl

package common;

require "$CFG::BASE/monitoring/config.cfg";
require "/usr/lib/cherry/dbaccess";

$tree = {};

$current_time = time;

#
# restoring monitor tree
#

sub restore_tree {
	my $mode = shift;

	my $profile_id = shift if $mode eq 'indirect';

	my $sth = $dbh->prepare(
		$mode eq 'indirect' ?
		"SELECT monitor_views.path, monitor_tree.type, monitor_tree.current_value, monitor_tree.current_e_level, NULL, NULL FROM monitor_tree INNER JOIN monitor_views ON monitor_views.element = monitor_tree.path WHERE monitor_views.profile_id = $profile_id UNION SELECT monitor_views.path, NULL, NULL, NULL, monitor_labels.label, FALSE FROM monitor_views INNER JOIN monitor_labels ON monitor_views.element = monitor_labels.path WHERE monitor_views.profile_id = $profile_id" :
		"SELECT path, type, current_value, current_e_level, NULL, NULL FROM monitor_tree UNION SELECT path, NULL, NULL, NULL, label, no_alarm FROM monitor_labels");
	$sth->execute;

	while(my($path, $type, $value, $e_level, $label, $no_alarm) =
		$sth->fetchrow_array) {
		my $pointer = $tree;
		my $walk_path = "";

		foreach my $node (split '\.', $path) {
			if(exists $pointer->{$node}) {
				$pointer = $pointer->{$node};
			} else {
				$pointer = $pointer->{$node} = {};
			}
			$walk_path = $walk_path ? "$walk_path.$node" : $node;
			$pointer->{visible} = ${"main::visible:$walk_path"} ? 1 : undef;
		}

		$pointer->{label} = $label;
		$pointer->{selected} = defined ${"main::selected:$path"} ? 1 : undef;

		unless($label or defined $no_alarm) {
			$pointer->{type} = $type;
			$pointer->{value} = $value;
			$pointer->{e_level} = $e_level;
		} else {
			$pointer->{no_alarm} = 1 if $no_alarm;
		}
	}

	return $tree unless wantarray;

	($last_message) = $dbh->selectrow_array("SELECT timestamp FROM monitor_messages ORDER BY timestamp DESC LIMIT 1");

	return ($tree, $last_message);
}

#
# smart sort (respect numbers)
#

sub smart_sort {
	my($a, $b) = @_;

#	if(my($prefixa, $suffixa) = ($a =~ /^(.*?)(\d+)$/) and my($prefixb, $suffixb) = ($b =~ /^(.*?)(\d+)$/)) {
#		if($prefixa eq $prefixb) {
#			return $suffixa <=> $suffixb;
#		}
#	}

#	return $a cmp $b;

	@a = split /(\D+)/, $a;
	shift @a if not defined $a[0];
	$len_a = scalar @a;

	@b = split /(\D+)/, $b;
	shift @b if not defined $b[0];
	$len_b = scalar @b;

	$len = $len_a < $len_b ? $len_a : $len_b;

	for($i = 0; $i < $len; $i++) {
		my $a = $a[$i];
		my $b = $b[$i];

		$r = ($a =~ /\d/ and $b =~ /\d/) ? $a <=> $b : $a cmp $b;

		return $r if $r;
	}

	return $a cmp $b;
}

#
# traverse monitor tree and store selected paths and attributes
#

sub traverse_tree {
	my($node, $selected, $visible, $path) = @_;

	next unless ref($node) eq 'HASH';

	if($node->{selected}) {
		push @$selected, [$path, $node->{type}];
	}

	if($node->{visible}) {
		push @$visible, $path;
	}

	foreach my $name (keys %$node) {
		my $subpath = $path ? "$path.$name" : $name;

		if(my $subnode = $node->{$name}) {
			traverse_tree($subnode, $selected, $visible, $subpath);
		}
	}
}

#
# find index of given element in array
#

sub find_index {
	my($item, $arr) = @_;

	my $index = 0;

	foreach(@$arr) {
		return $index if $_->[0] eq $item;
		$index++;
	}

	return undef;
}

#
# prefixen
#

$prefixen = {
	12	=> 'tera',
	9	=> 'giga',
	6	=> 'mega',
	3	=> 'kilo',
	0	=> '',
	-3	=> 'milli',
	-6	=> 'micro',
	-9	=> 'nano',
	-12	=> 'pico'
};

#
# data types, their programmatic and textual descriptions
#

$data_types = {
	BYTECOUNTER => {
		descr => 'byte counter',
		units => 'bytes',
		graph_units => 'bytes/s',
		type => 'COUNTER32'
	},
	STRING => {
		descr => 'just a string',
		units => 'string',
		type => 'STRING'
	},
	STATUS => {
		descr => 'status (up, down, administratively down)',
		units => 'up/down',
		type => 'TRIGGER',
		states => {
			up => 1,
			errors => .5,
			down => 0,
			'administratively down' => 0.1
		}
	},
	STATE => {
		descr => 'state (on, off)',
		units => 'on/off',
		type => 'TRIGGER',
		states => {
			on => 1,
			off => 0
		}
	},
	TEMPERATURE => {
		units => 'degrees',
		type => 'GAUGE',
	},
	SIGNALLEVEL => {
		descr => 'Level of signal',
		units => 'decibels',
		type => 'GAUGE',
	},
};

sub average {
	my($compound, $value, $amount) = @_;
	return $compound + $value * $amount;
}

sub min {
	my($compound, $value, $amount) = @_;
	return (not defined $compound or $value < $compound ? $value : $compound);
}

sub max {
	my($compound, $value, $amount) = @_;
	return (not defined $compound or $value > $compound ? $value : $compound);
}

$types = {
	TRIGGER => {
		process_graph_value => sub {
			my($type, $prev_value, $prev_timestamp, $value, $timestamp) = @_;

			my $states = $data_types->{$type}->{states};

			return $states->{$prev_value};
		},

		compound => \&min,
		not_scalable => 1
	},
	COUNTER32 => {
		process_graph_value => sub {
			my($type, $prev_value, $prev_timestamp, $value, $timestamp) = @_;
			if(defined $value and defined $prev_value) {
				my $diff = $value - $prev_value;
				$diff += 2 ** 32 if $diff < 0;

				return undef if $timestamp == $prev_timestamp;
				
				return $diff / ($timestamp - $prev_timestamp);
			}

			return undef;

		},
		process_value => sub {
			my($prev_value, $prev_timestamp, $value, $timestamp, $summary) = @_;

			return undef unless defined $value;
	
			my $retval = 'N/A';

			if(defined $prev_value) {
					$retval = $value - $prev_value;

					if($retval < 0) {
						$retval += 2 ** 32
					}				

					$summary->{summ} += $retval;

					$retval = reverse $retval;
					$retval =~ s<(\d\d\d)(?=\d)(?!\d*\.)><$1\.>g;
					$retval = reverse $retval;
			}
			return $retval;
		},
		process_summary => sub {
			my($summary) = @_;
			my $summ = $summary->{summ};
			$summ = reverse $summ;
			$summ =~ s<(\d\d\d)(?=\d)(?!\d*\.)><$1\.>g;
			$summ = reverse $summ;
			return (
				['', $summ]
			);
		}
	}
}
