#!/usr/bin/perl

# Network State Collector
# $Id: monitor.pl,v 2.0.0.1 2005/11/25 09:26:16 rz Exp $

use Socket;
use Symbol;
use Data::Dumper;
use Fcntl;

require "/etc/cherry/cherry.cfg";
require "$CFG::BASE/monitoring/config.cfg";

# server  config
my $bind_addr = $config::server_bind_addr;
my $port = $config::server_bind_port;
my $prompt = $config::server_prompt;
my $logfile = $config::log_file;

sub logmsg;
sub drop_connection;
sub cmd_shutdown;
sub restore_tree;
sub check_recency;

# connecting to database
require "/usr/lib/cherry/dbaccess";

# establishing server socket
my $proto = getprotobyname("tcp");
socket(SERVER, PF_INET, SOCK_STREAM, $proto) || die "socket: $!";
setsockopt(SERVER, SOL_SOCKET, SO_REUSEADDR,  pack("l", 1))   || die "setsockopt: $!";
setsockopt(SERVER, SOL_SOCKET, SO_KEEPALIVE,  pack("l", 1))   || die "setsockopt: $!";
my $addr = gethostbyname($bind_addr) or die "gethostbyname: $!";
bind(SERVER, sockaddr_in($port, inet_aton($bind_addr))) || die "bind: $!";
listen(SERVER, SOMAXCONN) || die;

logmsg "NSC server started on port $port";

$|=1;

my $fdset = '';

my %clients;

my $tree = {};

# how often we should check for pending action
my $sleep_time = $config::server_sleep_time;

# how often send server_alive messages
my $server_alive_period = $config::server_alive_period;

my $last_alive_message = time - $server_alive_period;

my $recency_check_interval = $config::recency_check_interval;

my $last_recency_check = time;

# our server socket
vec($fdset, fileno(SERVER), 1) = 1;

# on SIGHUP - shutdown
$SIG{HUP} = $SIG{INT} = $SIG{TERM} = sub {
	logmsg("got interrupt signal, shutting down");
	cmd_shutdown;
};

$SIG{PIPE} = 'IGNORE';

# restore monitoring tree from database
restore_tree;

my $global_time;

# here we go
while(1) {
	my($rin);
	my $nfound = select($rin = $fdset, undef, undef, $sleep_time);

	$global_time = time;

	if($last_recency_check <= $global_time - $recency_check_interval) {

		logmsg "checking recency";

		$dbh->begin_work;
		check_recency;
		$dbh->commit;
		$last_recency_check += $recency_check_interval;
	}

	if($last_alive_message <= $global_time - $server_alive_period) {
		logmsg "sending alive message";

		my $query = q|INSERT INTO monitor_messages(id, timestamp, path, type, value) VALUES (nextval('monitor_messages_id_seq'), ?, ?, ?, ?)|;
		$dbh->do($query, undef, $global_time, undef, 'SERVER', 'NSC server alive');
		$last_alive_message += $server_alive_period;
	}

	next if not $nfound;

	if(vec($rin, fileno(SERVER), 1)) {
		# new connection
		my $new_client = gensym;

		my $paddr = accept($new_client, SERVER);
		my($port,$iaddr) = sockaddr_in($paddr);
		my $name = gethostbyaddr($iaddr,AF_INET);

		logmsg "connection from $name [", inet_ntoa($iaddr), "] at port $port";

		# store his filehandle
		my $fd = fileno($new_client);
		$clients{$fd}->{fh} = $new_client;

		# autoflush
		set_autoflush($new_client, 1);
		fcntl($new_client, F_SETFL, O_NONBLOCK);
	
		# add descriptor into read mask
		vec($fdset, $fd, 1) = 1;

		print $new_client <<EOF;
You are connected to Network State Collector.
Enter NSC command to perform (use "list" to get a list of available commands).

EOF
		print $new_client $prompt;
	} 

	$dbh->begin_work;

	foreach my $fd (keys %clients) {
		if(vec($rin, $fd, 1)) {
			my $fh = $clients{$fd}->{fh};
			my $old = $clients{$fd}->{buf};

			my($recv, $buf, $rc);
	
			while( $rc = sysread($fh, $buf, 1024) > 0) { $recv .= $buf }

			logmsg "got from client: $recv";

			if($recv eq '') {
				drop_connection($fh);
				next;
			}

			$recv = $old . $recv;

			while($recv =~ /^(.*?)\n(.*)$/s) {
				my $cmd = $1;
				$recv = $2;
				my $response = process_cmd($cmd, $fh);

				print $fh $response;
			}

			logmsg "buffered: $recv";

			if(exists $clients{$fd}) {
				$clients{$fd}->{buf} = $recv;
			}
		}
	}

	$dbh->commit;
}

sub process_cmd {
	my($input, $fh) = @_;
	$input =~ s/[\r\n]//g;
	my($CMD, @ARGS) = split / /, $input;

	logmsg("got command $CMD");

	my($rc);

	if ($CMD =~ /^shutdown/i) { $rc = cmd_shutdown($fh, @ARGS); }

	elsif($CMD =~ /^[Ss]/) { $rc = cmd_set(@ARGS); }

	elsif ($CMD =~ /^[Gg]/) { $rc = cmd_get(@ARGS); }

	elsif ($CMD =~ /^[Ll]/) { $rc = cmd_list(@ARGS); }

	elsif ($CMD =~ /^[Qq]/) { $rc = cmd_quit($fh, @ARGS); }

	elsif ($CMD =~ /^[Dd]/) { $rc = cmd_dump(@ARGS); }
	elsif ($CMD =~ /^t/i) { $rc = "OK $global_time\n"; }

	elsif ($CMD =~ /^$/) { $rc = ""; }

	else { 
		$rc = "ERROR: Unknown command: $CMD\n"; 
	}

	return $rc.$prompt;
}

sub state_recent {
	my($pointer, $path, $time) = @_;

	if(ref($pointer) eq 'HASH') {
		foreach my $key (keys %$pointer) {
			state_recent($pointer->{$key}, "$path.$key", $time);
		}
		if(defined $pointer->{value}) {
			$pointer->{recent_till} = $time;
		}
	}
}

sub cmd_set {
	my @params = @_;
	my($rc, $mib, $pointer, @path);

	# Unimportant is default
	my $e_level = undef;

	my $path = shift @params;
	my $period = shift @params;
	if($params[0] =~ /^-e(\d+)$/) {
		$e_level = $1;
		shift @params;
	}
	my($type, @value) = @params;

	my $value = defined $value[0] ? join ' ', @value : undef;

	@path = split('\.', $path);

	unless(@path) {
		return "OK No op\n";
	}

	$pointer = $tree;

	$type = uc($type);

	foreach $mib (@path) {
		if(defined($pointer->{$mib})) {
			$pointer = $pointer->{$mib};
		} else {
			$pointer = $pointer->{$mib} = {}
		}
	}

	my $recent_till = $period ? $period + $global_time : undef;

	unless($type) {
		$rc = "OK $path is recent till $recent_till\n";
	} elsif($pointer->{type} ne $type or $pointer->{value} ne $value) {
		logmsg "writing new value: '$value', old one was '".$pointer->{value}."'";
		$pointer->{type} = $type;	
		$pointer->{value} = $value;	
		$pointer->{e_level} = $e_level;

		# write update message
		my $query = q|INSERT INTO monitor_messages(id, timestamp, path, type, value, e_level) VALUES (nextval('monitor_messages_id_seq'), ?, ?, ?, ?, ?)|;

		$dbh->do($query, undef, $global_time, $path, $type, $value, $e_level);

		# update database tree
		$query = q|UPDATE monitor_tree SET type = ?, current_value = ?, current_e_level = ? WHERE path = ?|;
		my($rows) = $dbh->do($query, undef, $type, $value, $e_level, $path);

		# no rows found
		if($rows == 0) {
			$query = q|INSERT INTO monitor_tree (path, date, type, current_value, current_e_level) VALUES (?, ?, ?, ?, ?)|;
			$dbh->do($query, undef, $path, $global_time, $type, $value, $e_level);

		}

		$rc = "OK $path = $type $value, recent till $recent_till\n";
	}

	if(defined $recent_till) {
		state_recent($pointer, $path, $recent_till);
	} else {
		delete $pointer->{recent_till};
	}

	return $rc;
}

sub cmd_get {
	my($object_id, $option) = @_;
	my($rc);
	my($mib);
	my($pointer);
	my(@path);

	@path = split('\.', $object_id);

	$pointer = $tree;

	foreach $mib (@path) {
		if(defined($pointer->{$mib})) {
			$pointer = $pointer->{$mib};
		} else {
			$rc = "ERROR: cannot find object \"$mib\" in the tree\n";
			return $rc;
		}
	}

	if($option =~ /leaves/) {
		my $mibs;
		foreach (keys %$pointer) { 
				if(!($_ eq "value" or $_ eq "type")) {$mibs.= " $_";}
		}
		return "OK $mibs\n";
	} elsif($option =~ /type/) {
		return "OK ".$pointer->{type}."\n";
	} elsif($option =~ /count/ || !defined($pointer->{value})) {
		my $mibs = 0;
		foreach (keys %$pointer) { 
				if(!($_ eq "value" or $_ eq "type")) {$mibs++;}
		}
		return "OK $mibs\n";
	} else {
		return "OK ".$pointer->{value}."\n";
	}
}

	
sub debug_dump_tree {
	my($level, $pointer, $spacer) = @_;
	my $result;

	unless(ref $pointer) {
		if(defined $pointer) {
			$result .= " = $pointer\n";
		} else {
			$result .= " = undef\n";
		}
	} else {
		$result .= "\n";
		foreach my $key ( sort { ($a =~ /\D/ or $b =~ /\D/) ? $a cmp $b : $a <=> $b } keys %$pointer) {
			$result .= "$level$key".dump_tree($level.$spacer, $pointer->{$key}, $spacer);
		}
	}

	return $result;
}

sub dump_tree {
	my($level, $pointer, $spacer) = @_;
	my $result;

	if(my $type = $pointer->{type}) {
		my $value = $pointer->{value};
		my $e_level = $pointer->{e_level};
		$value = 'UNDEFINED' if not defined $value;
		$result .= " $type: $value ($e_level)";
	}

	$result .= "\n";
	foreach my $key ( sort { ($a =~ /\D/ or $b =~ /\D/) ? $a cmp $b : $a <=> $b } keys %$pointer) {
		if(ref $pointer->{$key}) {
			$result .= "$level$key".dump_tree($level.$spacer, $pointer->{$key}, $spacer);
		}
	}

	return $result;
}

sub cmd_dump {
	my($object_id, $option) = @_;
	my($rc);
	my($mib);
	my($pointer);
	my(@path);

	@path = split('\.', $object_id);

	$pointer = $tree;

	foreach $mib (@path) {
		if(defined($pointer->{$mib})) {
			$pointer = $pointer->{$mib};
		} else {
			$rc = "ERROR: cannot find object \"$mib\" in the tree\n";
			return $rc;
		}
	}

	my $result;
	my $spacer = "  ";

	return dump_tree("", $pointer, $spacer)."\n";
}


	

sub cmd_list {
	
	my($rc) = <<EOF;
OK
There're the following NSC commands available:

SET <object_id> <period> [ [-eN] <type> <value>] - sets <object_id> value to <value> of type <type>, or states that subtree is recent <period> seconds. N is emergency level (0 - ok, 1 - warning, 2 - critical), if -e is not specified, value is unimportant
GET <object_id> [option] - reads the value of <object_id> using modifier [option]. If [option] is "type" returns <object_id> type
TIME - returns current system time
DUMP - dumps current tree state
QUIT - quit NSC session
SHUTDOWN - shutdown NSC server

EOF

	return $rc;
}

sub cmd_quit {
	my $fh = shift;	

	print $fh "OK Goodbye\n";

	logmsg "client quit";

	drop_connection($fh);
}

sub cmd_shutdown {
	my $fh = shift;
	logmsg "got shutdown command; shutting down";

	if($fh) {
		print $fh "OK Shutting down\n";
	}

	# dropping all connections
	foreach my $client(values %clients) {
		drop_connection($client->{fh});
	}

	# dropping server socket
	close SERVER;

	# disconnecting from database
	$dbh->disconnect;
	
	# shutting down
	exit;
}

sub check_recency {
	my($pointer, $path) = @_;
	unless (defined $pointer) { $pointer = $tree; $path = ''; }

	if(ref $pointer eq 'HASH') {
		if(defined $pointer->{value} and $pointer->{recent_till} < $global_time) {
			# recency period expired
			cmd_set($path, undef, $pointer->{type});
		}

		foreach my $key (keys %$pointer) {
			if(defined $pointer->{$key}) {
				check_recency($pointer->{$key}, $path ? "$path.$key" : $key);
			}
		}
	}
}

# retrieve monitoring tree from database
sub restore_tree {
	my $query = q|SELECT path, type, current_value, current_e_level FROM monitor_tree|;
	my $sth = $dbh->prepare($query);
	$sth->execute;
	while(my($path, $type, $value, $e_level) = $sth->fetchrow_array) {
		my $pointer = $tree;

		foreach my $key (split '\.', $path) {
			if(defined($pointer->{$key})) {
				$pointer = $pointer->{$key};
			} else {
				$pointer = $pointer->{$key} = {}
			}
		}

		$pointer->{type} = $type;
		$pointer->{value} = $value;
		$pointer->{e_level} = $e_level;
	}
}

sub set_autoflush {
	my($fh, $autoflush) = @_;
	my $old_fh = select;
	select($fh);
	$| = $autoflush;
	select($old_fh);
}

sub drop_connection {
	my $fh = shift;
	my $paddr = getpeername($fh);

	if($paddr) {
		my($port,$iaddr) = unpack_sockaddr_in($paddr);
		my $name = gethostbyaddr($iaddr,AF_INET);

		logmsg "connection from $name [", inet_ntoa($iaddr), "] at port $port droppped";
	} else {
		logmsg "fucked up paddr";
	}

	vec($fdset, fileno($fh), 1) = 0;
	delete $clients{fileno($fh)};
	close $fh;
}

# log procedure
sub logmsg { 
	print STDERR time.":".(scalar localtime).": @_\n";

	if(-f $logfile) {
		open(LOG, ">>$logfile");
		print LOG time.":".(scalar localtime).": @_\n";
		close(LOG);
	}
}
